unit DOT;
// Discrete Orthogonal transformations

interface
uses Complex, Transformation;

//----------- Fast Fourier Transformation Interface ----------------------------
type
  TFFTData = array of TCplxNumber;
  TCplxNumber = Complex.TCplxNumber;
  TDCTIIData = TShapeData;
  TDOTMode = (dmFFT, dmDCTII, dmNone, dmDHYT);
  TDOTState = (dsClosed, dsReady, dsComputing, dsFail);
  TDOTError = (deOK, deNotEnoughMem, deBusy, deIncorrectMode, deOpened, deClosed, deIncorrectSize, deInputDataSizeError);
  TDOTDirection = (ddForward, ddInverse);


function Initialize(X,Y:word; DOTMode:TDOTMode):TDOTError;
function Finalize:TDOTError;
function Transform(Dir:TDOTDirection; FData:TFFTData; DData:TDCTIIData):TDOTError;
function GetX:word;
function GetY:word;
function GetMode:TDOTMode;

var
  State:TDOTState;


implementation


//------------------- FFT Implementation ---------------------------------------

//internal variables
var
  N1,N2, halfN1, halfN2 : word;
  PowW1, PowW2 : TFFTData;
  Mirrors1, Mirrors2 : array of word;
  N1N2, halfN1N2 : longword;
  midPowW1, midPowW2 : array[-1..1] of TCplxNumber;
  Sqrt2 : real;
  DCTNorma : real;
  FFTData:TFFTData;
  DCTMode, IDCTMode : boolean;
  DHYTMode : boolean;
  Mode:TDOTMode;
  sqrtN1N2 : TFloat;


// Kernel of FFT (bit inversion and computations)
procedure FFTKernel(Data:TFFTData);
var blocksize,halfblock,block,expstep,exponent,ItLimit:word;
    x1,x2:TCplxNumber;
    i,j,item1,item2:Longword;
begin
  //If Inverse DCT II mode is on, multiply input with W^-1 coeficients
  if IDCTMode then begin
    for i:=0 to N1-1 do begin
      exponent:=(N1-i) mod N1; //compute original position (before mirroring)
      x1:=CplxMul(midPowW1[-(exponent and 1)],PowW1[(N1-(exponent shr 1)) mod N1]); //compute weighting coeficient W^-k
      j:=i;
      while j<N1N2 do begin //Multiply all rows with W
        Data[j]:=CplxMul(Data[j],x1);
        inc(j,N1);
      end;
    end;
  end;

  //bit inverse input data
  for j:=0 to N2-1 do
    for i:=0 to N1-1 do
      if (Mirrors1[i]<>65535) and (Mirrors2[j]<>65535) then begin
        if Mirrors1[i]<>i then begin
          x1:=Data[j*N1+i];
          Data[j*N1+i]:=Data[Mirrors2[j]*N1+Mirrors1[i]];
          Data[Mirrors2[j]*N1+Mirrors1[i]]:=x1;
        end;
        if Mirrors2[j]<>j then begin
          x1:=Data[j*N1+Mirrors1[i]];
          Data[j*N1+Mirrors1[i]]:=Data[Mirrors2[j]*N1+i];
          Data[Mirrors2[j]*N1+i]:=x1;
        end;
      end;

  // process lines
  j:=0;
  while j<N1N2 do begin //do for all lines
    asm
      mov blocksize,2
      mov ax,N1
      shr ax,1
      mov expstep,ax
    end;
    while blocksize<=N1 do begin
      asm
        mov ax,blocksize
        shr ax,1
        mov halfblock,ax
        mov block,0
      end;
      while block<N1 do begin  //do for all blocks
        item1:=j+block; item2:=j+block+halfblock;
        exponent:=0;
        while exponent<halfN1 do begin //for all blockitems
          {$o-}
          x2:=CplxMul(PowW1[exponent],Data[item2]);
          x1:=CplxAdd(Data[item1],x2);
          x2:=CplxSub(Data[item1],x2);
          {o+}
          Data[item1]:=x1;
          Data[item2]:=x2;
          inc(item1); inc(item2);
          inc(exponent,expstep);
        end;
        inc(block,blocksize);
      end;
      asm shl blocksize,1; shr expstep,1; end;
    end;
    inc(j,N1);
  end;
  //if Forward DCT II mode, then multiply results with wighting coeficient, and reset imagine parts
  if DCTMode then begin
    for i:=0 to halfN1-1 do begin
      x1:=CplxMul(PowW1[i shr 1],midPowW1[i and 1]); //compute wighting coeficient W^k
      j:=i;
      while j<N1N2 do begin //Multiply all rows
        Data[j]:=CplxMul(Data[j],x1);
        Data[j].i:=0;
        inc(j,N1);
      end;
    end;
  end;
  //for DHYT mode - prepare DHYT Data from FFT
  if DHYTMode then
    for j:=0 to N1N2-1 do begin
      Data[j].r:=Data[j].r-Data[j].i;
      Data[j].i:=0;
    end;


  //if Inverse DCT II mode is on, multiply input columns with weighting coeficient W^-k
  if IDCTMode then begin
    for j:=0 to N2-1 do begin
      item1:=Mirrors2[j]; //compute position before bit inversion
      if item1<>65535 then begin
        item2:=(N2-item1) mod N2; //compute original position before bit inversion and mirroring
        x1:=CplxMul(midPowW2[-(item2 and 1)],PowW2[(N2-item2 shr 1) mod N2]); //compute weighting coeficient from orig. position
        for i:=0 to halfN1-1 do begin //multiply all columns
          Data[j*N1+i].i:=0;
          Data[j*N1+i]:=CplxMul(Data[j*N1+i],x1);
        end;
        if item1<>j then begin //if mirroring was not identity
          item2:=(N2-j) mod N2; //compute original position for pair of j
          x1:=CplxMul(midPowW2[-(item2 and 1)],PowW2[(N2-item2 shr 1) mod N2]); //compute wighting coeficient for the pair
          for i:=0 to halfN1-1 do begin //multiply all columns
            Data[item1*N1+i].i:=0;
            Data[item1*N1+i]:=CplxMul(Data[item1*N1+i],x1);
          end;
        end;
      end;
    end;
  end;

  //process columns
  if DCTMode or IDCTMode then itLimit:=halfN1-1 else itLimit:=N1-1;
  for j:=0 to itLimit do begin  //do for all columns
    asm
      mov blocksize,2
      mov ax,N2
      shr ax,1
      mov expstep,ax
    end;
    while blocksize<=N2 do begin
      asm
        mov ax,blocksize
        shr ax,1
        mov halfblock,ax
        mov block,0
      end;
      while block<N2 do begin  //do for all blocks
        item1:=block*N1+j; item2:=(block+halfblock)*N1+j;
        exponent:=0;
        while exponent<halfN2 do begin //for all blockitems
          x2:=CplxMul(PowW2[exponent],Data[item2]);
          x1:=CplxAdd(Data[item1],x2);
          x2:=CplxSub(Data[item1],x2);
          Data[item1]:=x1;
          Data[item2]:=x2;
          inc(item1,N1); inc(item2,N1);
          inc(exponent,expstep);
        end;
        inc(block,blocksize);
      end;
      asm shl blocksize,1; shr expstep,1; end;
    end;
  end;
  //if DCT II mode is on, multiply result with weighting coeficent
  if DCTMode then begin
    for j:=0 to halfN2-1 do begin
      x1:=CplxMul(PowW2[j shr 1],midPowW2[j and 1]); //compute coeficient
      for i:=0 to halfN1-1 do begin
        Data[j*N1+i]:=CplxMul(Data[j*N1+i],x1);
        Data[j*N1+i].i:=0;
      end;
    end;
  end;
end;

// Returns bit inversion of "a" by "size"
function BinMirror(a,size:word):word;
asm
  shr size,1
  mov @result,0
@loop1:
  shr a,1
  rcl @result,1
  shr size,1
  jnz @loop1
end;

// Inititalize W1, W2, and Mirrors1, Mirrors2
function InitFFT(X,Y:word):boolean;
var i,m:word;
begin
  try
  N1:=X; N2:=Y;  // set size
  N1N2:=N1*N2; halfN1:=N1 shr 1; halfN2:=N2 shr 1; halfN1N2:=N1N2 shr 1;
  sqrtN1N2:=sqrt(N1N2);

  // create powers of W
  SetLength(PowW1,N1);
  for i:=0 to N1-1 do begin
    PowW1[i].r:=cos(2*Pi*i/N1);
    PowW1[i].i:=-sin(2*Pi*i/N1);
  end;

  SetLength(PowW2,N2);
  for i:=0 to N2-1 do begin
    PowW2[i].r:=cos(2*Pi*i/N2);
    PowW2[i].i:=-sin(2*Pi*i/N2);
  end;

  // create mirrors of indexes
  SetLength(Mirrors1,N1);
  for i:=0 to N1-1 do Mirrors1[i]:=N1;
  for i:=0 to N1-1 do
    if Mirrors1[i]=N1 then begin
      m:=BinMirror(i,N1);
      Mirrors1[i]:=m;
      if i<>m then Mirrors1[m]:=65535;
    end;

  SetLength(Mirrors2,N2);
  for i:=0 to N2-1 do Mirrors2[i]:=N2;
  for i:=0 to N2-1 do
    if Mirrors2[i]=N2 then begin
      m:=BinMirror(i,N2);
      Mirrors2[i]:=m;
      if i<>m then Mirrors2[m]:=65535;
    end;

  DCTMode:=false; IDCTMode:=false;
  DHYTMode:=false;
  Result:=true;
  except
    SetLength(PowW1,0); PowW1:=nil;
    SetLength(PowW2,0); PowW2:=nil;
    SetLength(Mirrors1,0); Mirrors1:=nil;
    SetLength(Mirrors2,0); Mirrors2:=nil;
    Result:=false;
  end;
end;


// Release allocated memory
procedure CloseFFT;
begin
  SetLength(PowW1,0); PowW1:=nil;
  SetLength(PowW2,0); PowW2:=nil;
  SetLength(Mirrors1,0); Mirrors1:=nil;
  SetLength(Mirrors2,0); Mirrors2:=nil;
end;


// Forward FFT
procedure FFT(Data:TFFTData);
var i : longword;
begin
  if N1=0 then exit;
  FFTKernel(Data);
  for i:=0 to N1N2-1 do begin
    Data[i].r:=Data[i].r/sqrtN1N2;
    Data[i].i:=Data[i].i/sqrtN1N2;
  end;
end;


// Backward (inverse) FFT
procedure IFFT(Data:TFFTData);
var i,j:longword;
    pomnum:TCplxNumber;
begin
  if N1=0 then exit;
  // prepare data for inverse transformation
  for i:=1 to (N1-1) shr 1 do begin  //mirror 1st line
    pomnum:=Data[i];
    Data[i]:=Data[N1-i];
    Data[N1-i]:=pomnum;
  end;
  for i:=1 to (N1-1) shr 1 do begin  //mirror middle line
    pomnum:=Data[halfN1N2+i];
    Data[halfN1N2+i]:=Data[halfN1N2+N1-i];
    Data[halfN1N2+N1-i]:=pomnum;
  end;
  for j:=1 to (N2-1) shr 1 do begin
    for i:=0 to (N1-1) shr 1 do
      if i>0 then begin
        pomnum:=Data[j*N1+i];
        Data[j*N1+i]:=Data[N1N2-j*N1+N1-i];
        Data[N1N2-j*N1+N1-i]:=pomnum;
        pomnum:=Data[j*N1+N1-i];
        Data[j*N1+N1-i]:=Data[N1N2-j*N1+i];
        Data[N1N2-j*N1+i]:=pomnum;
      end else begin  // Mirror 1st column
        pomnum:=Data[j*N1];
        Data[j*N1]:=Data[N1N2-j*N1];
        Data[N1N2-j*N1]:=pomnum;
      end;
      pomnum:=Data[j*N1+halfN1]; //Mirror middle column
      Data[j*N1+halfN1]:=Data[N1N2-j*N1+halfN1];
      Data[N1N2-j*N1+halfN1]:=pomnum;
    end;
  FFTKernel(Data);
  if Mode<>dmDCTII then
    for i:=0 to N1N2-1 do begin
      Data[i].r:=Data[i].r/sqrtN1N2;
      Data[i].i:=Data[i].i/sqrtN1N2;
    end;
end;
//------------------ End of FFT Implementation ---------------------------------




//------------------ DCT II Implementation -------------------------------------

function InitDCTII(X,Y:word):boolean;
begin
  try
  Result:=InitFFT(X shl 1,Y shl 1);
  midPowW1[0]:=powW1[0];
  midPowW1[1].r:=cos(Pi/N1); midPowW1[1].i:=-sin(Pi/N1);
  midPowW1[-1].r:=midPowW1[1].r; midPowW1[-1].i:=-midPowW1[1].i;
  midPowW2[0]:=powW2[0];
  midPowW2[1].r:=cos(Pi/N2); midPowW2[1].i:=-sin(Pi/N2);
  midPowW2[-1].r:=midPowW2[1].r; midPowW2[-1].i:=-midPowW2[1].i;
  SetLength(FFTData,N1N2);
  DCTNorma:=2/sqrt(halfN1*halfN2);
  except
  SetLength(FFTData,0);
  Result:=false;
  end;
end;

procedure DCTII(Data:TDCTIIData);
var i : word;
    j,j2 : longword;
begin
  j:=0; j2:=0;
  while j<halfN1N2 do begin
    for i:=0 to halfN1-1 do begin
      FFTData[j+i].r:=Data[j2+i];
      FFTData[j+i].i:=0;
      FFTData[j+i+halfN1].r:=0;
      FFTData[j+i+halfN1].i:=0;
      FFTData[j+halfN1N2+i].r:=0;
      FFTData[j+halfN1N2+i].i:=0;
      FFTData[j+halfN1N2+i+halfN1].r:=0;
      FFTData[j+halfN1N2+i+halfN1].i:=0;
    end;
    inc(j,N1); inc(j2,halfN1);
  end;
  DCTMode:=true;
  FFTKernel(FFTData);
  DCTMode:=false;
  j:=0; j2:=0;
  while j<halfN1N2 do begin
    for i:=0 to halfN1-1 do
      Data[j2+i]:=DCTNorma*FFTData[j+i].r;
    inc(j,N1); inc(j2,halfN1);
  end;
  for i:=0 to halfN1-1 do Data[i]:=Data[i]/Sqrt2;
  j:=0;
  for i:=0 to halfN2-1 do begin
    Data[j]:=Data[j]/Sqrt2;
    inc(j,halfN1);
  end;
end;

procedure IDCTII(Data:TDCTIIData);
var i : word;
    j,j2,j3 : longword;
begin
  j:=0; j2:=0;
  for j3:=0 to halfN2-1 do begin
    for i:=0 to halfN1-1 do begin
      FFTData[j+i].r:=DCTNorma*Data[j2+i];
      FFTData[j+i].i:=0;
      FFTData[j+i+halfN1].r:=0;
      FFTData[j+i+halfN1].i:=0;
      FFTData[j+halfN1N2+i].r:=0;
      FFTData[j+halfN1N2+i].i:=0;
      FFTData[j+halfN1N2+i+halfN1].r:=0;
      FFTData[j+halfN1N2+i+halfN1].i:=0;
    end;
    inc(j,N1); inc(j2,halfN1);
  end;
  for i:=0 to halfN1-1 do begin
    FFTData[i].r:=FFTData[i].r/Sqrt2;
    FFTData[i].i:=FFTData[i].i/Sqrt2;
  end;
  j:=0;
  for i:=0 to halfN2-1 do begin
    FFTData[j].r:=FFTData[j].r/Sqrt2;
    FFTData[j].i:=FFTData[j].i/Sqrt2;
    inc(j,N1);
  end;
  IDCTMode:=true;
  IFFT(FFTData);
  IDCTMode:=false;
  for j:=0 to halfN2-1 do
    for i:=0 to halfN1-1 do
      Data[j*halfN1+i]:=FFTData[j*N1+i].r;
end;

procedure CloseDCTII;
begin
  SetLength(FFTData,0); FFTData:=nil;
  CloseFFT;
end;
//---------------- End of DCT II Implementation --------------------------------



//------------------ DHYT Implementation ---------------------------------------

function InitDHYT(X,Y:word):boolean;
begin
  try
    Result:=InitFFT(X,Y);
    SetLength(FFTData,N1N2);
  except
    SetLength(FFTData,0);
    Result:=false;
  end;
end;

procedure DHYT(Data:TDCTIIData);
var
    j : longword;
begin
  for j:=0 to N1N2-1 do begin
    FFTData[j].r:=Data[j];
    FFTData[j].i:=0;
  end;
  DHYTMode:=true;
  FFT(FFTData);
  DHYTMode:=false;
  for j:=0 to N1N2-1 do
    Data[j]:=FFTData[j].r-FFTData[j].i;
end;

procedure IDHYT(Data:TDCTIIData);
begin
  DHYT(Data);
end;

procedure CloseDHYT;
begin
  CloseFFT;
  SetLength(FFTData,0); FFTData:=nil;
end;
//---------------- End of DCT II Implementation --------------------------------


function Bin1(a:word):byte;
asm
  mov cx,16
  mov Result,0
@loop1:
  shr a,1
  jnc  @jmp1
  inc Result
@jmp1:
  loopnz @loop1
end;

function Initialize(X,Y:word; DOTMode:TDOTMode):TDOTError;
begin
  case State of
   dsReady, dsFail : begin Result:=deOpened; exit end;
   dsComputing : begin Result:=deBusy; exit; end;
  else
    if (Bin1(X)<>1) or (Bin1(Y)<>1) then begin Result:=deIncorrectSize; exit; end;
    case DOTMode of
      dmNone: begin Result:=deIncorrectMode; exit; end;
      dmFFT: begin
        if (X=0) or (Y=0) or (X>32768) or (Y>32768) then begin Result:=deIncorrectSize; exit; end;
        if not InitFFT(X,Y) then begin Result:=deNotEnoughMem; exit end;
        Mode:=DOTMode;
        State:=dsReady;
        Result:=deOK;
      end;
      dmDCTII : begin
        if (X=0) or (Y=0) or (X>16384) or (Y>16384) then begin Result:=deIncorrectSize; exit; end;
        if not InitDCTII(X,Y) then begin Result:=deNotEnoughMem; exit end;
        Mode:=DOTMode;
        State:=dsReady;
        Result:=deOK;
      end;
      dmDHYT : begin
        if (X=0) or (Y=0) or (X>32768) or (Y>32768) then begin Result:=deIncorrectSize; exit; end;
        if not InitDHYT(X,Y) then begin Result:=deNotEnoughMem; exit end;
        Mode:=DOTMode;
        State:=dsReady;
        Result:=deOK;
      end;
      else Result:=deIncorrectMode;
    end;
  end;
end;

function Finalize:TDOTError;
begin
  case State of
   dsClosed : begin Result:=deClosed; exit end;
   dsComputing : begin Result:=deBusy; exit end;
  else
    case Mode of
     dmFFT: begin
       CloseFFT;
       Mode:=dmNone;
       State:=dsClosed;
       Result:=deOK;
     end;
     dmDCTII: begin
       CloseDCTII;
       Mode:=dmNone;
       State:=dsClosed;
       Result:=deOK;
     end;
     dmDHYT: begin
       CloseDHYT;
       Mode:=dmNone;
       State:=dsClosed;
       Result:=deOK;
     end;
     else Result:=deIncorrectMode;
    end;
  end;
end;

function GetX:word;
begin
  Result:=N1;
end;

function GetY:word;
begin
  Result:=N2;
end;

function Transform(Dir:TDOTDirection; FData:TFFTData; DData:TDCTIIData):TDOTError;
begin
  case State of
    dsClosed: begin Result:=deClosed; exit; end;
    dsComputing: begin Result:=deBusy; exit; end;
  else
    State:=dsComputing;
    case Mode of
      dmFFT: begin
        if (FData=nil) and (DData<>nil) then begin State:=dsFail; Result:=deIncorrectMode; exit end;
        if Length(FData)<>N1N2 then begin State:=dsFail; Result:=deInputDataSizeError; exit; end;
        case Dir of
          ddForward: FFT(FData);
          ddInverse: IFFT(FData);
        end;
      end;
      dmDCTII: begin
        if (FData<>nil) and (DData=nil) then begin State:=dsFail; Result:=deIncorrectMode; exit end;
        if Length(DData)<>halfN1N2 shr 1 then begin State:=dsFail; Result:=deInputDataSizeError; exit; end;
        case Dir of
          ddForward: DCTII(DData);
          ddInverse: IDCTII(DData);
        end;
      end;
      dmDHYT: begin
        if (FData<>nil) and (DData=nil) then begin State:=dsFail; Result:=deIncorrectMode; exit end;
        if Length(DData)<>N1N2 then begin State:=dsFail; Result:=deInputDataSizeError; exit; end;
        case Dir of
          ddForward: DHYT(DData);
          ddInverse: IDHYT(DData);
        end;
      end;
    end;
    State:=dsReady;
    Result:=deOK;
  end;
end;

function GetMode:TDOTMode;
begin
  Result:=Mode;
end;

begin
  N1:=0;
  Sqrt2:=Sqrt(2);
  Mode:=dmNone;
  State:=dsClosed;
end.
